"""extensions
Example:
    
    from flask_sqlalchemy import SQLAlchemy
    db = SQLAlchemy()


Usage:

    # in settings
    EXTENSIONS = [
        ...
        'api.extensions.db'
        ...
    ]
"""



from flask import g, request

from flask_mico.error import InvalidParameterError
from flask_mico.log import logger


def on_json_loading_failed(e):
    raise InvalidParameterError(
        "Failed to decode JSON object: {0}".format(e))


class Translate:
    """Translate
        - application/json
    """

    def process_request(self):
        logger.debug("Translate - method: %s, content_type: %s",
                     repr(request.method), repr(request.content_type))
        if not request.is_json:
            return
        request.on_json_loading_failed = on_json_loading_failed
        raw_json = request.get_json()
        if not isinstance(raw_json, dict):
            raise InvalidParameterError(
                "A valid JSON document is required")
        g.data = raw_json



"""errors
Example:

1. 
    from enum import Enum
    class APIError(Enum):
        SESSION_EXPIRED = {
            "status": 400,
            "code": 1001,
            "message": "session has expired"
        }
2. 
    SESSION_EXPIRED = {
            "status": 400,
            "code": 1001,
            "message": "session has expired"
    }

Usage:

    ...
    from flask_mico.error import AppError
    class Index(ApiView):
        def get(self):
            # raise AppError(SESSION_EXPIRED) or /
            raise AppError(APIError.SESSION_EXPIRED)
    ...
"""



from importlib import import_module
from typing import List


def import_class_in_module(mod_path: str):
    try:
        mod = import_module(mod_path)
    except ModuleNotFoundError:
        mod_path, cls = mod_path.rsplit('.', 1)
    mod = import_module(mod_path)
    return getattr(mod, cls)


def register_extensions(app, extensions: List[str]):
    for ext_path in extensions:
        ext = import_class_in_module(ext_path)
        ext.init_app(app)



from importlib import import_module
from types import ModuleType
from urllib.parse import urljoin

from flask_mico import ApiView
from flask_mico.log import logger


def include(mod_path: str):
    return import_module(mod_path)


def handle_app_urls(prefix: str, urls_mod):
    for url in urls_mod.urlpatterns:
        if not issubclass(url[1], ApiView):
            logger.error("'%s' Not an instance of 'ApiView'", str(url[1]))
            break
        path = urljoin(prefix, url[0])
        logger.debug('Route %s', path)
        yield path, url[1]


def register_routes(app, urls_path: str):
    def add_url_rule(path, api_cls):
        app.add_url_rule(path, view_func=api_cls.as_view(api_cls.__name__))

    urls_mod = import_module(urls_path)
    logger.debug('register routes %s', urls_mod.urlpatterns)
    for url_ in urls_mod.urlpatterns:
        if isinstance(url_[1], ApiView):
            add_url_rule(url_[0], url_[1])
        elif isinstance(url_[1], ModuleType):
            for path, view in handle_app_urls(url_[0], url_[1]):
                add_url_rule(path, view)



from functools import wraps
from typing import Dict

from cerberus import SchemaError, Validator
from flask import g, request

from flask_mico.error import AppError, InvalidParameterError
from flask_mico.log import logger


def validator(schema: Dict,
              in_params: bool = False):
    """validator: validate data according to `schema`
    Args:
        schema: schema to validate data
        in_params: when is_params is true, validate `request.args` data

    Usage:
        @validator(schema=ALLSCHEMA["User"], is_params=True)
        def get(self): 
            ...
    """
    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            schema_validator(schema, in_params)
            return func(*args, **kwargs)
        return wrapper
    return decorate


def schema_validator(schema: Dict,
                     in_params: bool = False):
    """validator: validate data according to `schema`
    Args:
        schema: schema to validate data
        in_params: when is_params is true, validate `request.args` data
    """
    _validator = Validator(schema)
    if in_params:
        doc_data = dict(request.args)
    else:
        # This is the translated `g.data` from the `flask_mico.middlewares.Translate`
        doc_data = g.get("data", {})
    logger.debug("FieldValidator - data: %s", repr(doc_data))
    try:
        if not _validator.validate(doc_data):
            raise InvalidParameterError(_validator.errors)
    except SchemaError:
        raise InvalidParameterError(
            description="FieldValidator Schema Error")



"""
setup logger
"""
import logging
import os
from logging.handlers import RotatingFileHandler

FORMAT = '[%(asctime)s] [%(process)d] [%(levelname)s] %(message)s'
TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S %z'


logger = logging.getLogger("TestMicroserviceAPI")
logger.propagate = False


def setup_logger(level, logdir, service_name):
    # process level
    if isinstance(level, bool) and level is True:
        level = logging.DEBUG
    else:
        level = logging.INFO
    # clear logging default handlers
    logging.getLogger().handlers.clear()
    logger.handlers.clear()
    # add logger format
    formatter = logging.Formatter(FORMAT, TIMESTAMP_FORMAT)
    # set logger handler
    if level == logging.INFO:
        # add rotaing file handler
        log_file_name = 'logger-%s.log' % service_name
        log_file_str = os.path.join(logdir, log_file_name)
        file_handler = RotatingFileHandler(
            filename=log_file_str,
            maxBytes=1024 * 1024 * 50,
            backupCount=5, encoding='utf-8')
        file_handler.suffix = "_%Y-%m-%d_%H.log"
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    # default console handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.NOTSET)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    # set level
    logger.setLevel(level)



from flask import jsonify, request
from flask.views import MethodView

from flask_mico import error
from flask_mico.constant import MEDIA_JSON, http_status_message

try:
    from collections import OrderedDict
except ImportError:
    OrderedDict = dict


class ApiView(MethodView):
    def on_error(self, status, msg=None, data=None):
        body = OrderedDict()
        body["code"] = status
        body["success"] = False
        body["message"] = msg or http_status_message(status)
        body["data"] = data
        return jsonify(body), status, {"Content-Type": MEDIA_JSON}

    def on_success(self, status=200, msg=None, data=None):
        body = OrderedDict()
        body["code"] = status
        body["success"] = True
        body["message"] = msg or http_status_message(status)
        body["data"] = data
        return jsonify(body), status, {"Content-Type": MEDIA_JSON}

    def get(self, *args, **kwargs):
        raise error.NotSupportedError(method="GET", url=request.path)

    def post(self, *args, **kwargs):
        raise error.NotSupportedError(method="POST", url=request.path)

    def put(self, *args, **kwargs):
        raise error.NotSupportedError(method="PUT", url=request.path)

    def delete(self, *args, **kwargs):
        raise error.NotSupportedError(method="DELETE", url=request.path)



import os

SERVICE_NAME = '{{project_name}}'
DOMAIN = os.getenv("DOMAIN", default="example.com")

ALLOW_METHODS = ("GET", "POST", "PUT", "DELETE", "OPTIONS")
BASE_DIR = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
LOG_DIR = os.path.join(BASE_DIR, 'logs')

MIDDLEWARES = [
    'flask_mico.middleware.Translate'
]

EXTENSIONS = [
]

ROOT_URLCONF = '{{project_name}}.urls'



import json
from flask import Flask
from werkzeug.exceptions import HTTPException

from flask_mico.constant import MEDIA_JSON, http_status_message
from flask_mico.log import logger

from .errors import AppError, HTTPStatus

try:
    from collections import OrderedDict
except:
    OrderedDict = dict


def _handle_bad_request(e):
    """handle bad request
    """
    body = OrderedDict()
    body["code"] = e.code
    body["message"] = getattr(
        e, 'description', http_status_message(e.code))
    body["success"] = False
    body["data"] = None
    resp = e.get_response()
    resp.content_type = MEDIA_JSON
    body = json.dumps(body)
    resp.data = body
    logger.debug("HTTPException: status: %s body: %s",
                 e.code, body)
    return resp


def register_errors(app: Flask):
    """Register errors
    """
    app.register_error_handler(AppError, lambda e: e.handle())
    app.register_error_handler(HTTPStatus, lambda e: e.handle())
    app.register_error_handler(HTTPException, _handle_bad_request)



"""middlewares

Example:

    class MiddlewareExample:
        def process_request(self):
            # https://flask.palletsprojects.com/en/1.1.x/api/#flask.Flask.before_request
            
            # do_somethings
            print("MiddlewareExample", process_request)
        
        def process_response(self,resp):
            # https://flask.palletsprojects.com/en/1.1.x/api/#flask.Flask.after_request
            
            # do_somethings
            print("MiddlewareExample", process_response)
            return resp

Usage:

    # in settings
    MIDDLEWARES = [
        ...
        'api.middlewares.MiddlewareExample'
        ...
    ]
"""



import enum
from typing import Dict

from flask import Flask, jsonify, make_response

from flask_mico.constant import MEDIA_JSON, http_status_message
from flask_mico.log import logger

try:
    from collections import OrderedDict
except:
    OrderedDict = dict


# common errors
ERR_INVALID_PARAMETER = {
    "status": 400,
    "code": 400,
    "message": "Invalid Parameter"
}
ERR_AUTH_REQUIRED = {
    "status": 401,
    "code": 401,
    "message": "Authentication Required"
}
ERR_NOT_SUPPORTED = {
    "status": 404,
    "code": 404,
    "message": "Not Supported"
}
ERR_UNKNOWN = {
    "status": 500,
    "code": 500,
    "message": "Unknown Error"
}


class AppError(Exception):
    def __init__(self, error: Dict = ERR_UNKNOWN,
                 description: str = None, **kwargs):
        """
        Args:
            error: an instance of 'Dict' or 'Enum'
        """
        self._error = error
        if isinstance(error, enum.Enum):
            self._error = error.value
        self._kv = kwargs
        self._error["description"] = description

    @property
    def status(self) -> int:
        return self._error["status"]

    @property
    def code(self) -> int:
        return self._error["code"]

    @property
    def message(self) -> str:
        return self._error["message"]

    @property
    def description(self) -> str:
        return self._error["description"]

    def handle(self):
        body = OrderedDict()
        body["code"] = self.code
        body["message"] = self.message
        body["success"] = False
        body["data"] = self._kv.get("data", None)
        body_ = jsonify(body)
        resp = make_response(body_, self.status)
        resp.content_type = MEDIA_JSON
        logger.debug("AppError: status: %s body: %s, desc: %s",
                     self.code, body_, self.description)
        return resp


class InvalidParameterError(AppError):
    def __init__(self, description=None):
        super().__init__(ERR_INVALID_PARAMETER)
        self._error["description"] = description


class AuthenticationRequiredError(AppError):
    def __init__(self, description=None):
        super().__init__(ERR_AUTH_REQUIRED)
        self._error["description"] = description


class NotSupportedError(AppError):
    def __init__(self, method=None, url=None):
        super().__init__(ERR_NOT_SUPPORTED)
        if method and url:
            self._error["description"] = "method: %s, url: %s" % (method, url)


class HTTPStatus(Exception):
    def __init__(self, body: str, status: int, headers: Dict):
        self.body = body
        self.status = status
        self.headers = headers

    def handle(self):
        return make_response(self.body, self.status, self.headers)



"""field_schema
# https://docs.python-cerberus.org/en/stable/validation-rules.html

Example:
    
    INDEX_POST = {
        "hello": {
            "type": "string",
            "regex": r"^[0-9a-zA-Z_]{4,30}$",
            "required": True
        }
    }

Usage:

    ...
    from flask_mico import validator
    class Index(ApiView):
        @validator(schema=INDEX_POST)
        def get(self):
            return self.on_success(msg="pong", data={"pong": True})
    ...
"""



"""
common constant
"""
from werkzeug.http import HTTP_STATUS_CODES

MEDIA_JSON = 'application/json'


def http_status_message(code):
    """Maps an HTTP status code to the textual status"""
    return HTTP_STATUS_CODES.get(code, '')



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="flask_mico",
    version="0.1.7",
    description="flask mico",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="@buzz",
    url="https://github.com/YE-Kits/flask_mico",
    platforms=["all"],
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3"
    ],
    install_requires=[
        'flask',
        'cerberus'
    ],
    entry_points={
        'console_scripts': [
            'startproject=flask_mico.cli:create_project'
        ]
    }
)



from typing import Any, List

from flask import Flask

from flask_mico.util import import_class_in_module


def _handle(app: Flask, mid_: Any, point_name: str = None):
    """
    Args:
        app: a instance of Flask
        mid_: Any object
        point_name: default=None(apply in root app)
    """
    if hasattr(mid_, 'process_request'):
        if callable(mid_.process_request):
            app.before_request_funcs.setdefault(
                point_name, []).append(mid_.process_request)
    if hasattr(mid_, 'process_response'):
        if callable(mid_.process_response):
            app.after_request_funcs.setdefault(
                point_name, []).append(mid_.process_response)


def register_middlewares(app: Flask, middlewares: List[Any]):
    """register middlewares apply in app
    Args:
        app: a instance of Flask
        middlewares: List of mid in order
    """
    for mid_path in middlewares:
        mid_ = import_class_in_module(mid_path)
        _handle(app, mid_())



import os
import re
import sys

BASE_DIR = os.path.abspath((os.path.dirname(__file__)))
TEMLLATE_DIR = os.path.join(BASE_DIR, 'project_template')
TEM_ARG = "{{project_name}}"


def create_project():
    if len(sys.argv) < 2:
        print('Missing project name!')
        return
    name = sys.argv[1]
    if not re.findall(r"[a-zA-Z_]*", name):
        print('Project name unqualified!')
        return

    pro_path = "./"
    if len(sys.argv) >= 3:
        if not sys.argv[2].startswith('/'):
            pro_path = os.path.join(pro_path, sys.argv[2])
        else:
            pro_path = sys.argv[2]
    # make project templates
    for (root, dirs, files) in os.walk(TEMLLATE_DIR):
        if "__pycache__" in dirs:
            dirs.remove("__pycache__")
        mid_path = root.replace(TEMLLATE_DIR, '')
        mid_path = mid_path.replace(TEM_ARG, name).lstrip('/')
        if mid_path:
            path_ = os.path.join(pro_path, mid_path)
        else:
            path_ = pro_path

        for file in files:
            if not os.path.exists(path_):
                os.mkdir(path_)
            with open(os.path.join(root, file), 'r') as f:
                content = f.read().replace(TEM_ARG, name)
                wirte2file(os.path.join(path_, file), content)


def wirte2file(path, content):
    with open(path, 'w') as f:
        f.write(content)



from flask import Flask

from flask_mico.conf import settings
from flask_mico.error.err_handle import register_errors
from flask_mico.log import setup_logger
from flask_mico.middleware.mid_handle import register_middlewares
from flask_mico.util import register_extensions
from flask_mico.urls import register_routes


def create_app():
    app = Flask(__name__)
    app.config.from_object(settings)

    setup_logger(settings.DEBUG, settings.LOG_DIR, settings.SERVICE_NAME)
    register_errors(app)
    register_middlewares(app, middlewares=settings.MIDDLEWARES)
    register_extensions(app, settings.EXTENSIONS)
    register_routes(app, settings.ROOT_URLCONF)
    return app
